#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Read to and write from file with envelopes."""

import logging
from contextlib import contextmanager
from logging import Logger
from typing import IO, Optional, Union

from aea.configurations.base import PublicId
from aea.helpers import file_lock
from aea.helpers.base import exception_log_and_reraise
from aea.mail.base import Envelope


SEPARATOR = b","

_default_logger = logging.getLogger(__name__)


def _encode(e: Envelope, separator: bytes = SEPARATOR):
    result = b""
    result += e.to.encode("utf-8")
    result += separator
    result += e.sender.encode("utf-8")
    result += separator
    result += str(e.protocol_id).encode("utf-8")
    result += separator
    result += e.message_bytes
    result += separator

    return result


def _decode(e: bytes, separator: bytes = SEPARATOR):
    split = e.split(separator)

    if len(split) < 5 or split[-1] not in [b"", b"\n"]:
        raise ValueError(
            "Expected at least 5 values separated by commas and last value being empty or new line, got {}".format(
                len(split)
            )
        )

    to = split[0].decode("utf-8").strip().lstrip("\x00")
    sender = split[1].decode("utf-8").strip()
    protocol_id = PublicId.from_str(split[2].decode("utf-8").strip())
    # protobuf messages cannot be delimited as they can contain an arbitrary byte sequence; however
    # we know everything remaining constitutes the protobuf message.
    message = SEPARATOR.join(split[3:-1])
    # message = codecs.decode(message, "unicode-escape").encode("utf-8")  # noqa: E800

    return Envelope(to=to, sender=sender, protocol_id=protocol_id, message=message)


@contextmanager
def lock_file(file_descriptor: IO[bytes], logger: Logger = _default_logger):
    """Lock file in context manager.

    :param file_descriptor: file descriptio of file to lock.
    """
    with exception_log_and_reraise(
        logger.error, f"Couldn't acquire lock for file {file_descriptor.name}: {{}}",
    ):
        file_lock.lock(file_descriptor, file_lock.LOCK_EX)

    try:
        yield
    finally:
        file_lock.unlock(file_descriptor)


def write_envelope(
    envelope: Envelope,
    file_pointer: IO[bytes],
    separator: bytes = SEPARATOR,
    logger: Logger = _default_logger,
) -> None:
    """Write envelope to file."""
    encoded_envelope = _encode(envelope, separator=separator)
    logger.debug("write {}: to {}".format(encoded_envelope, file_pointer.name))
    write_with_lock(file_pointer, encoded_envelope, logger)


def write_with_lock(
    file_pointer: IO[bytes], data: Union[bytes], logger: Logger = _default_logger
) -> None:
    """Write bytes to file protected with file lock."""
    with lock_file(file_pointer, logger):
        file_pointer.write(data)
        file_pointer.flush()


def envelope_from_bytes(
    bytes_: bytes, separator: bytes = SEPARATOR, logger: Logger = _default_logger
) -> Optional[Envelope]:
    """
    Decode bytes to get the envelope.

    :return: Envelope
    :raise: Exception
    """
    logger.debug("processing: {!r}".format(bytes_))
    envelope = None  # type: Optional[Envelope]
    try:
        envelope = _decode(bytes_, separator=separator)
    except ValueError as e:
        logger.error("Bad formatted input: {!r}. {}".format(bytes_, e))
    except Exception as e:  # pragma: nocover # pylint: disable=broad-except
        logger.exception("Error when processing a input. Message: {}".format(str(e)))
    return envelope



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests for the aea configurations project."""
import os
import shutil
import tempfile
from unittest.mock import Mock

import pytest

from aea.cli import cli
from aea.configurations.project import AgentAlias, Project
from aea.helpers.base import cd
from aea.test_tools.click_testing import CliRunner

from tests.conftest import MAX_FLAKY_RERUNS, MY_FIRST_AEA_PUBLIC_ID, ROOT_DIR


class TestProjectAndAgentAlias:
    """Check project and agent alias."""

    def setup(self):
        """Set the test up."""
        self.cwd = os.getcwd()
        self.t = tempfile.mkdtemp()
        os.chdir(self.t)
        self.runner = CliRunner()
        self.project_public_id = MY_FIRST_AEA_PUBLIC_ID
        self.project_path = os.path.join(
            self.t, self.project_public_id.author, self.project_public_id.name
        )

    def _test_project(self, is_local: bool, skip_consistency_check: bool):
        """Test method to handle both local and remote registry."""
        registry_path = os.path.join(ROOT_DIR, "packages")
        project = Project.load(
            self.t,
            self.project_public_id,
            is_local=is_local,
            registry_path=registry_path,
            skip_consistency_check=skip_consistency_check,
        )
        assert os.path.exists(self.project_path)

        with cd(self.project_path):
            result = self.runner.invoke(
                cli,
                ["--skip-consistency-check", "config", "get", "agent.agent_name"],
                catch_exceptions=False,
                standalone_mode=False,
            )
        assert self.project_public_id.name in result.output
        project.remove()
        assert not os.path.exists(self.project_path)

    def test_project_local(self):
        """Test project loaded and removed, from local registry."""
        self._test_project(True, False)

    @pytest.mark.integration
    @pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
    def test_project_remote(self):
        """Test project loaded and removed, from remove registry."""
        self.project_public_id = MY_FIRST_AEA_PUBLIC_ID.to_latest()
        self._test_project(False, True)

    def test_agents(self):
        """Test agent added to project and rmeoved."""
        project = Project(self.project_public_id, self.project_path)
        alias = AgentAlias(project, "test", [], Mock(), Mock())
        assert project.agents
        alias.remove_from_project()
        assert not project.agents
        assert all(key in alias.dict for key in ["agent_name", "config", "public_id"])

    def teardown(self):
        """Tear dowm the test."""
        os.chdir(self.cwd)
        try:
            shutil.rmtree(self.t)
        except (OSError, IOError):
            pass



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the resources class."""
from contextlib import suppress
from typing import List, Optional, cast

from aea.components.base import Component
from aea.configurations.base import ComponentId, ComponentType, PublicId
from aea.connections.base import Connection
from aea.contracts.base import Contract
from aea.protocols.base import Protocol
from aea.registries.base import (
    AgentComponentRegistry,
    ComponentRegistry,
    HandlerRegistry,
    Registry,
)
from aea.skills.base import Behaviour, Handler, Model, Skill


class Resources:
    """This class implements the object that holds the resources of an AEA."""

    def __init__(self, agent_name: str = "standalone") -> None:
        """
        Instantiate the resources.

        :return None
        """
        self._agent_name = agent_name
        self._component_registry = AgentComponentRegistry(agent_name=agent_name)
        self._handler_registry = HandlerRegistry(agent_name=agent_name)
        self._behaviour_registry = ComponentRegistry[Behaviour](agent_name=agent_name)
        self._model_registry = ComponentRegistry[Model](agent_name=agent_name)

        self._registries = [
            self._component_registry,
            self._handler_registry,
            self._behaviour_registry,
            self._model_registry,
        ]  # type: List[Registry]

    @property
    def agent_name(self) -> str:
        """Get the agent name."""
        return self._agent_name

    @property
    def component_registry(self) -> AgentComponentRegistry:
        """Get the agent component registry."""
        return self._component_registry

    @property
    def behaviour_registry(self) -> ComponentRegistry[Behaviour]:
        """Get the behaviour registry."""
        return self._behaviour_registry

    @property
    def handler_registry(self) -> HandlerRegistry:
        """Get the handler registry."""
        return self._handler_registry

    @property
    def model_registry(self) -> ComponentRegistry[Model]:
        """Get the model registry."""
        return self._model_registry

    def add_component(self, component: Component):
        """Add a component to resources."""
        if component.component_type == ComponentType.PROTOCOL:
            self.add_protocol(cast(Protocol, component))
        elif component.component_type == ComponentType.SKILL:
            self.add_skill(cast(Skill, component))
        elif component.component_type == ComponentType.CONTRACT:
            self.add_contract(cast(Contract, component))
        elif component.component_type == ComponentType.CONNECTION:
            self.add_connection(cast(Connection, component))
        else:
            raise ValueError(
                "Component type {} not supported.".format(
                    component.component_type.value
                )
            )

    def add_protocol(self, protocol: Protocol) -> None:
        """
        Add a protocol to the set of resources.

        :param protocol: a protocol
        :return: None
        """
        self._component_registry.register(protocol.component_id, protocol)

    def get_protocol(self, protocol_id: PublicId) -> Optional[Protocol]:
        """
        Get protocol for given protocol id.

        :param protocol_id: the protocol id
        :return: a matching protocol, if present, else None
        """
        protocol = self._component_registry.fetch(
            ComponentId(ComponentType.PROTOCOL, protocol_id)
        )
        return cast(Protocol, protocol)

    def get_all_protocols(self) -> List[Protocol]:
        """
        Get the list of all the protocols.

        :return: the list of protocols.
        """
        protocols = self._component_registry.fetch_by_type(ComponentType.PROTOCOL)
        return cast(List[Protocol], protocols)

    def remove_protocol(self, protocol_id: PublicId) -> None:
        """
        Remove a protocol from the set of resources.

        :param protocol_id: the protocol id for the protocol to be removed.
        :return: None
        """
        self._component_registry.unregister(
            ComponentId(ComponentType.PROTOCOL, protocol_id)
        )

    def add_contract(self, contract: Contract) -> None:
        """
        Add a contract to the set of resources.

        :param contract: a contract
        :return: None
        """
        self._component_registry.register(contract.component_id, contract)

    def get_contract(self, contract_id: PublicId) -> Optional[Contract]:
        """
        Get contract for given contract id.

        :param contract_id: the contract id
        :return: a matching contract, if present, else None
        """
        contract = self._component_registry.fetch(
            ComponentId(ComponentType.CONTRACT, contract_id)
        )
        return cast(Contract, contract)

    def get_all_contracts(self) -> List[Contract]:
        """
        Get the list of all the contracts.

        :return: the list of contracts.
        """
        contracts = self._component_registry.fetch_by_type(ComponentType.CONTRACT)
        return cast(List[Contract], contracts)

    def remove_contract(self, contract_id: PublicId) -> None:
        """
        Remove a contract from the set of resources.

        :param contract_id: the contract id for the contract to be removed.
        :return: None
        """
        self._component_registry.unregister(
            ComponentId(ComponentType.CONTRACT, contract_id)
        )

    def add_connection(self, connection: Connection) -> None:
        """
        Add a connection to the set of resources.

        :param connection: a connection
        :return: None
        """
        self._component_registry.register(connection.component_id, connection)

    def get_connection(self, connection_id: PublicId) -> Optional[Connection]:
        """
        Get connection for given connection id.

        :param connection_id: the connection id
        :return: a matching connection, if present, else None
        """
        connection = self._component_registry.fetch(
            ComponentId(ComponentType.CONNECTION, connection_id)
        )
        return cast(Connection, connection)

    def get_all_connections(self) -> List[Connection]:
        """
        Get the list of all the connections.

        :return: the list of connections.
        """
        connections = self._component_registry.fetch_by_type(ComponentType.CONNECTION)
        return cast(List[Connection], connections)

    def remove_connection(self, connection_id: PublicId) -> None:
        """
        Remove a connection from the set of resources.

        :param connection_id: the connection id for the connection to be removed.
        :return: None
        """
        self._component_registry.unregister(
            ComponentId(ComponentType.CONNECTION, connection_id)
        )

    def add_skill(self, skill: Skill) -> None:
        """
        Add a skill to the set of resources.

        :param skill: a skill
        :return: None
        """
        self._component_registry.register(skill.component_id, skill)
        if skill.handlers is not None:
            for handler in skill.handlers.values():
                self._handler_registry.register(
                    (skill.public_id, handler.name), handler
                )
        if skill.behaviours is not None:
            for behaviour in skill.behaviours.values():
                self._behaviour_registry.register(
                    (skill.public_id, behaviour.name), behaviour
                )
        if skill.models is not None:
            for model in skill.models.values():
                self._model_registry.register((skill.public_id, model.name), model)

    def get_skill(self, skill_id: PublicId) -> Optional[Skill]:
        """
        Get the skill for a given skill id.

        :param skill_id: the skill id
        :return: a matching skill, if present, else None
        """
        skill = self._component_registry.fetch(
            ComponentId(ComponentType.SKILL, skill_id)
        )
        return cast(Skill, skill)

    def get_all_skills(self) -> List[Skill]:
        """
        Get the list of all the skills.

        :return: the list of skills.
        """
        skills = self._component_registry.fetch_by_type(ComponentType.SKILL)
        return cast(List[Skill], skills)

    def remove_skill(self, skill_id: PublicId) -> None:
        """
        Remove a skill from the set of resources.

        :param skill_id: the skill id for the skill to be removed.
        :return: None
        """
        self._component_registry.unregister(ComponentId(ComponentType.SKILL, skill_id))
        with suppress(ValueError):
            self._handler_registry.unregister_by_skill(skill_id)
        with suppress(ValueError):
            self._behaviour_registry.unregister_by_skill(skill_id)
        with suppress(ValueError):
            self._model_registry.unregister_by_skill(skill_id)

    def get_handler(
        self, protocol_id: PublicId, skill_id: PublicId
    ) -> Optional[Handler]:
        """
        Get a specific handler.

        :param protocol_id: the protocol id the handler is handling
        :param skill_id: the skill id of the handler's skill
        :return: the handler
        """
        handler = self._handler_registry.fetch_by_protocol_and_skill(
            protocol_id, skill_id
        )
        return handler

    def get_handlers(self, protocol_id: PublicId) -> List[Handler]:
        """
        Get all handlers for a given protocol.

        :param protocol_id: the protocol id the handler is handling
        :return: the list of handlers matching the protocol
        """
        handlers = self._handler_registry.fetch_by_protocol(protocol_id)
        return handlers

    def get_all_handlers(self) -> List[Handler]:
        """
        Get all handlers from all skills.

        :return: the list of handlers
        """
        handlers = self._handler_registry.fetch_all()
        return handlers

    def get_behaviour(
        self, skill_id: PublicId, behaviour_name: str
    ) -> Optional[Behaviour]:
        """
        Get a specific behaviours for a given skill.

        :param skill_id: the skill id
        :param behaviour_name: the behaviour name
        :return: the behaviour, if it is present, else None
        """
        behaviour = self._behaviour_registry.fetch((skill_id, behaviour_name))
        return behaviour

    def get_behaviours(self, skill_id: PublicId) -> List[Behaviour]:
        """
        Get all behaviours for a given skill.

        :param skill_id: the skill id
        :return: the list of behaviours of the skill
        """
        behaviours = self._behaviour_registry.fetch_by_skill(
            skill_id
        )  # type: List[Behaviour]
        return behaviours

    def get_all_behaviours(self) -> List[Behaviour]:
        """
        Get all behaviours from all skills.

        :return: the list of all behaviours
        """
        behaviours = self._behaviour_registry.fetch_all()
        return behaviours

    def setup(self) -> None:
        """
        Set up the resources.

        Calls setup on all resources.

        :return: None
        """
        for r in self._registries:
            r.setup()

    def teardown(self) -> None:
        """
        Teardown the resources.

        Calls teardown on all resources.

        :return: None
        """
        for r in self._registries:
            r.teardown()



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2020 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""AEA configuration utils."""
from functools import singledispatch
from typing import Dict, Optional, Set, Tuple

from aea.configurations.base import (
    AgentConfig,
    ComponentId,
    ComponentType,
    ConnectionConfig,
    ContractConfig,
    PackageConfiguration,
    ProtocolConfig,
    PublicId,
    SkillConfig,
)


@singledispatch
def replace_component_ids(
    _arg: PackageConfiguration,
    _replacements: Dict[ComponentType, Dict[PublicId, PublicId]],
):
    """
    Update public id references in a package configuration.

    This depends on the actual configuration being considered.
    """


@replace_component_ids.register(AgentConfig)  # type: ignore
def _(
    arg: AgentConfig, replacements: Dict[ComponentType, Dict[PublicId, PublicId]],
):
    """
    Replace references in agent configuration.

    It breaks down in:
    1) replace public ids in 'protocols', 'connections', 'contracts' and 'skills';
    2) replace public ids in default routing;
    3) replace public id of default connection;
    4) replace custom component configurations.

    :param arg: the agent configuration.
    :param replacements: the replacement mapping.
    :return: None
    """
    _replace_component_id(
        arg,
        {
            ComponentType.PROTOCOL,
            ComponentType.CONNECTION,
            ComponentType.CONTRACT,
            ComponentType.SKILL,
        },
        replacements,
    )

    # update default routing
    protocol_replacements = replacements.get(ComponentType.PROTOCOL, {})
    connection_replacements = replacements.get(ComponentType.CONNECTION, {})
    for protocol_id, connection_id in list(arg.default_routing.items()):

        # update protocol (if replacements provides it)
        new_protocol_id = protocol_replacements.get(protocol_id, protocol_id)
        old_value = arg.default_routing.pop(protocol_id)
        arg.default_routing[new_protocol_id] = old_value
        # in case needs to be used below
        protocol_id = new_protocol_id

        # update connection (if replacements provides it)
        new_connection_id = connection_replacements.get(connection_id, connection_id)
        arg.default_routing[protocol_id] = new_connection_id

    # update default connection
    if arg.default_connection is not None:
        default_connection_public_id = arg.default_connection
        new_default_connection_public_id = replacements.get(
            ComponentType.CONNECTION, {}
        ).get(default_connection_public_id, default_connection_public_id)
        arg.default_connection = new_default_connection_public_id

    for component_id in set(arg.component_configurations.keys()):
        replacements_by_type = replacements.get(component_id.component_type, {})
        if component_id.public_id in replacements_by_type:
            new_component_id = ComponentId(
                component_id.component_type,
                replacements_by_type[component_id.public_id],
            )
            old_value = arg.component_configurations.pop(component_id)
            arg.component_configurations[new_component_id] = old_value


@replace_component_ids.register(ProtocolConfig)  # type: ignore
def _(
    _arg: ProtocolConfig, _replacements: Dict[ComponentType, Dict[PublicId, PublicId]],
):
    """Do nothing - protocols have no references."""


@replace_component_ids.register(ConnectionConfig)  # type: ignore
def _(
    arg: ConnectionConfig, replacements: Dict[ComponentType, Dict[PublicId, PublicId]],
):
    """Replace references in a connection configuration."""
    _replace_component_id(
        arg, {ComponentType.PROTOCOL, ComponentType.CONNECTION}, replacements
    )

    protocol_replacements = replacements.get(ComponentType.PROTOCOL, {})
    for old_protocol_id in set(arg.restricted_to_protocols):
        new_protocol_id = protocol_replacements.get(old_protocol_id, old_protocol_id)
        arg.restricted_to_protocols.remove(old_protocol_id)
        arg.restricted_to_protocols.add(new_protocol_id)

    for old_protocol_id in set(arg.excluded_protocols):
        new_protocol_id = protocol_replacements.get(old_protocol_id, old_protocol_id)
        arg.excluded_protocols.remove(old_protocol_id)
        arg.excluded_protocols.add(new_protocol_id)


@replace_component_ids.register(ContractConfig)  # type: ignore
def _(  # type: ignore
    _arg: ContractConfig, _replacements: Dict[ComponentType, Dict[PublicId, PublicId]],
):
    """Do nothing - contracts have no references."""


@replace_component_ids.register(SkillConfig)  # type: ignore
def _(
    arg: SkillConfig, replacements: Dict[ComponentType, Dict[PublicId, PublicId]],
):
    """Replace references in a skill configuration."""
    _replace_component_id(
        arg,
        {
            ComponentType.PROTOCOL,
            ComponentType.CONNECTION,
            ComponentType.CONTRACT,
            ComponentType.SKILL,
        },
        replacements,
    )


def _replace_component_id(
    config: PackageConfiguration,
    types_to_update: Set[ComponentType],
    replacements: Dict[ComponentType, Dict[PublicId, PublicId]],
):
    """
    Replace a component id.

    :param config: the component configuration to update.
    :param types_to_update: the types to update.
    :param replacements: the replacements.
    :return:
    """
    for component_type in types_to_update:
        public_id_set = getattr(config, component_type.to_plural(), set())
        replacements_given_type = replacements.get(component_type, {})
        for old_public_id in list(public_id_set):
            new_public_id = replacements_given_type.get(old_public_id, old_public_id)
            public_id_set.remove(old_public_id)
            public_id_set.add(new_public_id)


def get_latest_component_id_from_prefix(
    agent_config: AgentConfig, component_prefix: Tuple[ComponentType, str, str]
) -> Optional[ComponentId]:
    """
    Get component id with the greatest version in an agent configuration given its prefix.

    :param agent_config: the agent configuration.
    :param component_prefix: the package prefix.
    :return: the package id with the greatest version, or None if not found.
    """
    all_dependencies = agent_config.package_dependencies
    chosen_component_ids = [
        c for c in all_dependencies if c.component_prefix == component_prefix
    ]
    nb_results = len(chosen_component_ids)
    return chosen_component_ids[0] if nb_results == 1 else None



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""Implementation of the 'aea launch' subcommand."""
import sys
from collections import OrderedDict
from pathlib import Path
from typing import List, cast

import click

from aea.cli.utils.click_utils import AgentDirectory
from aea.cli.utils.context import Context
from aea.cli.utils.loggers import logger
from aea.helpers.multiple_executor import ExecutorExceptionPolicies
from aea.launcher import AEALauncher


@click.command()
@click.argument("agents", nargs=-1, type=AgentDirectory())
@click.option("--multithreaded", is_flag=True)
@click.pass_context
def launch(click_context, agents: List[str], multithreaded: bool):
    """Launch many agents at the same time."""
    _launch_agents(click_context, agents, multithreaded)


def _launch_agents(
    click_context: click.core.Context, agents: List[str], multithreaded: bool
) -> None:
    """
    Run multiple agents.

    :param click_context: click context object.
    :param agents: agents names.
    :param multithreaded: bool flag to run as multithreads.

    :return: None.
    """
    agents_directories = list(map(Path, list(OrderedDict.fromkeys(agents))))
    mode = "threaded" if multithreaded else "multiprocess"
    ctx = cast(Context, click_context.obj)

    launcher = AEALauncher(
        agent_dirs=agents_directories,
        mode=mode,
        fail_policy=ExecutorExceptionPolicies.log_only,
        log_level=ctx.verbosity,
    )

    try:
        """
        run in threaded mode and wait for thread finished cause issue with python 3.6/3.7 on windows
        probably keyboard interrupt exception gets lost in executor pool or in asyncio module
        """
        launcher.start(threaded=True)
        launcher.join_thread()
    except KeyboardInterrupt:
        logger.info("Keyboard interrupt detected.")
    finally:
        launcher.stop()

    for agent in launcher.failed:
        logger.info(f"Agent {agent} terminated with exit code 1")

    for agent in launcher.not_failed:
        logger.info(f"Agent {agent} terminated with exit code 0")

    logger.debug(f"Exit cli. code: {launcher.num_failed}")
    sys.exit(1 if launcher.num_failed > 0 else 0)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This package contains the strategy model."""

from typing import Dict, List, Optional, Tuple, cast

from aea.crypto.ledger_apis import LedgerApis
from aea.helpers.transaction.base import Terms
from aea.skills.base import Model

from packages.fetchai.contracts.staking_erc20.contract import PUBLIC_ID
from packages.fetchai.skills.confirmation_aw1.registration_db import RegistrationDB


REQUIRED_KEYS = [
    "ethereum_address",
    "fetchai_address",
    "signature_of_ethereum_address",
    "signature_of_fetchai_address",
    "developer_handle",
]
DEFAULT_TOKEN_DISPENSE_AMOUNT = 100000
DEFAULT_TOKEN_DENOMINATION = "atestfet"  # nosec
DEFAULT_CONTRACT_ADDRESS = "0x351bac612b50e87b46e4b10a282f632d41397de2"
DEFAULT_OVERRIDE = False


class Strategy(Model):
    """This class is the strategy model."""

    def __init__(self, **kwargs) -> None:
        """
        Initialize the strategy of the agent.

        :return: None
        """
        self._token_denomination = kwargs.pop(
            "token_denomination", DEFAULT_TOKEN_DENOMINATION
        )
        self._token_dispense_amount = kwargs.pop(
            "token_dispense_amount", DEFAULT_TOKEN_DISPENSE_AMOUNT
        )
        self._fetchai_staking_contract_address = kwargs.pop(
            "fetchai_staking_contract_address", DEFAULT_CONTRACT_ADDRESS
        )
        self._override_staking_check = kwargs.pop(
            "override_staking_check", DEFAULT_OVERRIDE
        )
        self._aw2_aeas: List[str] = kwargs.pop("aw2_aeas", [])
        super().__init__(**kwargs)
        self._is_ready_to_register = False
        self._is_registered = False
        self.is_registration_pending = False
        self.signature_of_ethereum_address: Optional[str] = None
        self._ledger_id = self.context.default_ledger_id
        self._max_tx_fee = 100
        self._contract_ledger_id = "ethereum"
        self._contract_callable = "get_stake"
        self._contract_id = str(PUBLIC_ID)
        self._in_process_registrations: Dict[str, Dict[str, str]] = {}

    @property
    def contract_id(self) -> str:
        """Get the ledger on which the contract is deployed."""
        return self._contract_id

    @property
    def contract_address(self) -> str:
        """Get contract address."""
        return self._fetchai_staking_contract_address

    @property
    def contract_ledger_id(self) -> str:
        """Get the ledger on which the contract is deployed."""
        return self._contract_ledger_id

    @property
    def contract_callable(self) -> str:
        """Get the ledger on which the contract is deployed."""
        return self._contract_callable

    @property
    def aw2_aeas(self) -> List[str]:
        """Get list of AW2 AEAs."""
        return self._aw2_aeas

    @property
    def all_registered_aeas(self) -> List[str]:
        """Get all the registered AEAs."""
        registration_db = cast(RegistrationDB, self.context.registration_db)
        all_registered = registration_db.get_all_registered()
        return all_registered

    def lock_registration_temporarily(self, address: str, info: Dict[str, str]) -> None:
        """Lock this address for registration."""
        self._in_process_registrations.update({address: info})

    def finalize_registration(self, address: str) -> None:
        """Lock this address for registration."""
        info = self._in_process_registrations.pop(address)
        self.context.logger.info(
            f"finalizing registration for address={address}, info={info}"
        )
        registration_db = cast(RegistrationDB, self.context.registration_db)
        registration_db.set_registered(
            address=address,
            ethereum_address=info["ethereum_address"],
            ethereum_signature=info["signature_of_ethereum_address"],
            fetchai_signature=info["signature_of_fetchai_address"],
            developer_handle=info["developer_handle"],
            tweet=info.get("tweet", ""),
        )

    def unlock_registration(self, address: str) -> None:
        """Unlock this address for registration."""
        info = self._in_process_registrations.pop(address, {})
        self.context.logger.info(
            f"registration info did not pass staking checks = {info}"
        )

    def valid_registration(
        self, registration_info: Dict[str, str], sender: str
    ) -> Tuple[bool, int, str]:
        """
        Check if the registration info is valid.

        :param registration_info: the registration info
        :param sender: the sender
        :return: tuple of success, error code and error message
        """
        if not all([key in registration_info for key in REQUIRED_KEYS]):
            return (
                False,
                1,
                f"missing keys in registration info, required: {REQUIRED_KEYS}!",
            )
        if not sender == registration_info["fetchai_address"]:
            return (
                False,
                1,
                "fetchai address of agent and registration info do not match!",
            )
        if not self._valid_signature(
            registration_info["ethereum_address"],
            registration_info["signature_of_fetchai_address"],
            sender,
            "ethereum",
        ):
            return (False, 1, "fetchai address and signature do not match!")
        if not self._valid_signature(
            sender,
            registration_info["signature_of_ethereum_address"],
            registration_info["ethereum_address"],
            "fetchai",
        ):
            return (False, 1, "ethereum address and signature do not match!")
        if sender in self._in_process_registrations:
            return (False, 1, "registration in process for this address!")
        registration_db = cast(RegistrationDB, self.context.registration_db)
        if registration_db.is_registered(sender):
            return (False, 1, "already registered!")
        return (True, 0, "all good!")

    def _valid_signature(
        self, expected_signer: str, signature: str, message_str: str, ledger_id: str
    ) -> bool:
        """
        Check if the signature and message match the expected signer.

        :param expected_signer: the signer
        :param signature: the signature
        :param message_str: the message
        :param ledger_id: the ledger id
        :return: bool indiciating validity
        """
        try:
            result = expected_signer in LedgerApis.recover_message(
                ledger_id, message_str.encode("utf-8"), signature
            )
        except Exception as e:  # pylint: disable=broad-except
            self.context.logger.warning(f"Signing exception: {e}")
            result = False
        return result

    def get_terms(self, counterparty: str) -> Terms:
        """
        Get terms of transaction.

        :param counterparty: the counterparty to receive funds
        :return: the terms
        """
        terms = Terms(
            ledger_id=self._ledger_id,
            sender_address=self.context.agent_address,
            counterparty_address=counterparty,
            amount_by_currency_id={
                self._token_denomination: -self._token_dispense_amount
            },
            quantities_by_good_id={},
            is_sender_payable_tx_fee=True,
            nonce="some",
            fee_by_currency_id={self._token_denomination: self._max_tx_fee},
        )
        return terms

    @staticmethod
    def get_kwargs(info: Dict[str, str]) -> Dict[str, str]:
        """
        Get the kwargs for the contract state call.

        :param counterparty:
        """
        counterparty = info["ethereum_address"]
        return {"address": counterparty}

    def has_staked(self, state: Dict[str, str]) -> bool:
        """
        Check if the agent has staked.

        :return: bool, indicating outcome
        """
        if self._override_staking_check:
            return True
        result = int(state.get("stake", "0")) > 0
        return result



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This package contains a a behaviour to request data from a HTTP endpoint."""

import json
from typing import cast

from aea.skills.behaviours import TickerBehaviour

from packages.fetchai.connections.http_client.connection import (
    PUBLIC_ID as HTTP_CLIENT_PUBLIC_ID,
)
from packages.fetchai.protocols.http.message import HttpMessage
from packages.fetchai.skills.simple_data_request.dialogues import HttpDialogues


DEFAULT_REQUEST_INTERVAL = 20.0


class HttpRequestBehaviour(TickerBehaviour):
    """This class defines an http request behaviour."""

    def __init__(self, **kwargs):
        """Initialise the behaviour."""
        request_interval = kwargs.pop(
            "request_interval", DEFAULT_REQUEST_INTERVAL
        )  # type: int
        self.url = kwargs.pop("url", None)
        self.method = kwargs.pop("method", None)
        self.body = kwargs.pop("body", None)
        if self.url is None or self.method is None or self.body is None:
            raise ValueError("Url, method and body must be provided.")
        super().__init__(tick_interval=request_interval, **kwargs)

    def setup(self) -> None:
        """
        Implement the setup.

        :return: None
        """
        pass

    def act(self) -> None:
        """
        Implement the act.

        :return: None
        """
        self._generate_http_request()

    def _generate_http_request(self) -> None:
        """
        Generate http request to provided url with provided body and method.

        :return: None
        """
        http_dialogues = cast(HttpDialogues, self.context.http_dialogues)
        request_http_message, _ = http_dialogues.create(
            counterparty=str(HTTP_CLIENT_PUBLIC_ID),
            performative=HttpMessage.Performative.REQUEST,
            method=self.method,
            url=self.url,
            headers="",
            version="",
            body=json.dumps(self.body).encode("utf-8"),
        )
        self.context.outbox.put_message(message=request_http_message)

    def teardown(self) -> None:
        """
        Implement the task teardown.

        :return: None
        """
        pass



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests of the handler classes of the tac participation skill."""

import logging
from pathlib import Path
from typing import cast
from unittest.mock import patch

import pytest

from aea.protocols.dialogue.base import DialogueMessage
from aea.test_tools.test_skill import BaseSkillTestCase, COUNTERPARTY_ADDRESS

from packages.fetchai.protocols.oef_search.message import OefSearchMessage
from packages.fetchai.protocols.state_update.message import StateUpdateMessage
from packages.fetchai.protocols.tac.message import TacMessage
from packages.fetchai.skills.tac_participation.dialogues import (
    OefSearchDialogues,
    StateUpdateDialogue,
    StateUpdateDialogues,
    TacDialogues,
)
from packages.fetchai.skills.tac_participation.game import Game, Phase
from packages.fetchai.skills.tac_participation.handlers import (
    OefSearchHandler,
    TacHandler,
)

from tests.conftest import ROOT_DIR


class TestOefSearchHandler(BaseSkillTestCase):
    """Test oef search handler of tac participation."""

    path_to_skill = Path(ROOT_DIR, "packages", "fetchai", "skills", "tac_participation")

    @classmethod
    def setup(cls):
        """Setup the test class."""
        super().setup()
        cls.oef_search_handler = cast(
            OefSearchHandler, cls._skill.skill_context.handlers.oef
        )
        cls.game = cast(Game, cls._skill.skill_context.game)
        cls.oef_dialogues = cast(
            OefSearchDialogues, cls._skill.skill_context.oef_search_dialogues
        )
        cls.list_of_messages = (
            DialogueMessage(
                OefSearchMessage.Performative.SEARCH_SERVICES, {"query": "some_query"}
            ),
        )
        cls.controller_address = "some_controller_address"

    def test_setup(self):
        """Test the setup method of the oef handler."""
        assert self.oef_search_handler.setup() is None
        self.assert_quantity_in_outbox(0)

    def test_handle_unidentified_dialogue(self):
        """Test the _handle_unidentified_dialogue method of the oef handler."""
        # setup
        incorrect_dialogue_reference = ("", "")
        incoming_message = self.build_incoming_message(
            message_type=OefSearchMessage,
            dialogue_reference=incorrect_dialogue_reference,
            performative=OefSearchMessage.Performative.SEARCH_RESULT,
            to=self.skill.skill_context.agent_address
            + "_"
            + str(self.skill.skill_context.skill_id),
        )

        # operation
        with patch.object(self.oef_search_handler.context.logger, "log") as mock_logger:
            self.oef_search_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.WARNING,
            f"received invalid oef_search message={incoming_message}, unidentified dialogue.",
        )

    def test_handle_error(self):
        """Test the _handle_error method of the oef handler."""
        # setup
        oef_dialogue = self.prepare_skill_dialogue(
            dialogues=self.oef_dialogues, messages=self.list_of_messages[:1],
        )
        incoming_message = cast(
            OefSearchMessage,
            self.build_incoming_message_for_skill_dialogue(
                dialogue=oef_dialogue,
                performative=OefSearchMessage.Performative.OEF_ERROR,
                to=self.skill.skill_context.agent_address
                + "_"
                + str(self.skill.skill_context.skill_id),
                oef_error_operation=OefSearchMessage.OefErrorOperation.SEARCH_SERVICES,
            ),
        )

        # operation
        with patch.object(self.oef_search_handler.context.logger, "log") as mock_logger:
            self.oef_search_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.WARNING,
            f"received OEF Search error: dialogue_reference={oef_dialogue.dialogue_label.dialogue_reference}, oef_error_operation={incoming_message.oef_error_operation}",
        )

    def test_on_search_result(self):
        """Test the _on_search_result method of the oef handler."""
        # setup
        oef_dialogue = self.prepare_skill_dialogue(
            dialogues=self.oef_dialogues, messages=self.list_of_messages[:1],
        )
        incoming_message = cast(
            OefSearchMessage,
            self.build_incoming_message_for_skill_dialogue(
                dialogue=oef_dialogue,
                performative=OefSearchMessage.Performative.SEARCH_RESULT,
                to=self.skill.skill_context.agent_address
                + "_"
                + str(self.skill.skill_context.skill_id),
                agents=(self.controller_address,),
            ),
        )
        self.game._phase = Phase.PRE_GAME

        # operation
        with patch.object(self.oef_search_handler.context.logger, "log") as mock_logger:
            self.oef_search_handler.handle(incoming_message)

        # after
        self.assert_quantity_in_outbox(1)

        # _on_search_result
        mock_logger.assert_any_call(
            logging.DEBUG,
            f"on search result: dialogue_reference={oef_dialogue.dialogue_label.dialogue_reference} agents={incoming_message.agents}",
        )

        # _on_controller_search_result
        mock_logger.assert_any_call(
            logging.INFO, "found the TAC controller. Registering...",
        )

        # _register_to_tac
        assert self.game._expected_controller_addr == self.controller_address
        assert self.game.phase == Phase.GAME_REGISTRATION

        has_attributes, error_str = self.message_has_attributes(
            actual_message=self.get_message_from_outbox(),
            message_type=TacMessage,
            performative=TacMessage.Performative.REGISTER,
            to=self.controller_address,
            sender=self.skill.skill_context.agent_address,
            agent_name=self.skill.skill_context.agent_name,
        )
        assert has_attributes, error_str

        assert self.game._tac_dialogue is not None
        assert self.skill.skill_context.behaviours.tac_search.is_active is False

    def test_on_controller_search_result_i(self):
        """Test the _on_controller_search_result method of the oef handler where phase is not PRE_GAME."""
        # setup
        oef_dialogue = self.prepare_skill_dialogue(
            dialogues=self.oef_dialogues, messages=self.list_of_messages[:1],
        )
        incoming_message = cast(
            OefSearchMessage,
            self.build_incoming_message_for_skill_dialogue(
                dialogue=oef_dialogue,
                performative=OefSearchMessage.Performative.SEARCH_RESULT,
                to=self.skill.skill_context.agent_address
                + "_"
                + str(self.skill.skill_context.skill_id),
                agents=("agent_1", "agent_2"),
            ),
        )
        self.game._phase = Phase.GAME

        # operation
        with patch.object(self.oef_search_handler.context.logger, "log") as mock_logger:
            self.oef_search_handler.handle(incoming_message)

        # after
        # _on_search_result
        mock_logger.assert_any_call(
            logging.DEBUG,
            f"on search result: dialogue_reference={oef_dialogue.dialogue_label.dialogue_reference} agents={incoming_message.agents}",
        )

        # _on_controller_search_result
        mock_logger.assert_any_call(
            logging.DEBUG,
            "ignoring controller search result, the agent is already competing.",
        )

    def test_on_controller_search_result_ii(self):
        """Test the _on_controller_search_result method of the oef handler where list of agent addresses is empty."""
        # setup
        oef_dialogue = self.prepare_skill_dialogue(
            dialogues=self.oef_dialogues, messages=self.list_of_messages[:1],
        )
        incoming_message = cast(
            OefSearchMessage,
            self.build_incoming_message_for_skill_dialogue(
                dialogue=oef_dialogue,
                performative=OefSearchMessage.Performative.SEARCH_RESULT,
                to=self.skill.skill_context.agent_address
                + "_"
                + str(self.skill.skill_context.skill_id),
                agents=tuple(),
            ),
        )
        self.game._phase = Phase.PRE_GAME

        # operation
        with patch.object(self.oef_search_handler.context.logger, "log") as mock_logger:
            self.oef_search_handler.handle(incoming_message)

        # after
        # _on_search_result
        mock_logger.assert_any_call(
            logging.DEBUG,
            f"on search result: dialogue_reference={oef_dialogue.dialogue_label.dialogue_reference} agents={incoming_message.agents}",
        )

        # _on_controller_search_result
        mock_logger.assert_any_call(
            logging.INFO, "couldn't find the TAC controller. Retrying...",
        )

    def test_on_controller_search_result_more_than_one_agents(self):
        """Test the _on_controller_search_result method of the oef handler where list of agents contains more than one agents."""
        # setup
        oef_dialogue = self.prepare_skill_dialogue(
            dialogues=self.oef_dialogues, messages=self.list_of_messages[:1],
        )
        incoming_message = cast(
            OefSearchMessage,
            self.build_incoming_message_for_skill_dialogue(
                dialogue=oef_dialogue,
                performative=OefSearchMessage.Performative.SEARCH_RESULT,
                to=self.skill.skill_context.agent_address
                + "_"
                + str(self.skill.skill_context.skill_id),
                agents=("agent_1", "agent_2"),
            ),
        )
        self.game._phase = Phase.PRE_GAME

        # operation
        with patch.object(self.oef_search_handler.context.logger, "log") as mock_logger:
            self.oef_search_handler.handle(incoming_message)

        # after
        # _on_search_result
        mock_logger.assert_any_call(
            logging.DEBUG,
            f"on search result: dialogue_reference={oef_dialogue.dialogue_label.dialogue_reference} agents={incoming_message.agents}",
        )

        # _on_controller_search_result
        mock_logger.assert_any_call(
            logging.WARNING, "found more than one TAC controller. Retrying...",
        )

    def test_handle_invalid(self):
        """Test the _handle_invalid method of the oef handler."""
        # setup
        invalid_performative = OefSearchMessage.Performative.UNREGISTER_SERVICE
        incoming_message = self.build_incoming_message(
            message_type=OefSearchMessage,
            dialogue_reference=("1", ""),
            performative=invalid_performative,
            to=self.skill.skill_context.agent_address
            + "_"
            + str(self.skill.skill_context.skill_id),
            service_description="some_service_description",
        )

        # operation
        with patch.object(self.oef_search_handler.context.logger, "log") as mock_logger:
            self.oef_search_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.WARNING,
            f"cannot handle oef_search message of performative={invalid_performative} in dialogue={self.oef_dialogues.get_dialogue(incoming_message)}.",
        )

    def test_teardown(self):
        """Test the teardown method of the oef_search handler."""
        assert self.oef_search_handler.teardown() is None
        self.assert_quantity_in_outbox(0)


class TestTacHandler(BaseSkillTestCase):
    """Test tac handler of tac participation."""

    path_to_skill = Path(ROOT_DIR, "packages", "fetchai", "skills", "tac_participation")

    @classmethod
    def setup(cls):
        """Setup the test class."""
        super().setup()
        cls.tac_handler = cast(TacHandler, cls._skill.skill_context.handlers.tac)
        cls.tac_dialogues = cast(TacDialogues, cls._skill.skill_context.tac_dialogues)
        cls.state_update_dialogues = cast(
            StateUpdateDialogues, cls._skill.skill_context.state_update_dialogues
        )
        cls.game = cast(Game, cls._skill.skill_context.game)
        cls.logger = cls.tac_handler.context.logger

        cls.agent_name = "some_agent_name"
        cls.amount_by_currency_id = {"1": 10}
        cls.exchange_params_by_currency_id = {"1": 1.0}
        cls.quantities_by_good_id = {"2": 10}
        cls.utility_params_by_good_id = {"2": 1.0}
        cls.fee_by_currency_id = {"1": 1}
        cls.agent_addr_to_name = {COUNTERPARTY_ADDRESS: "some_name"}
        cls.currency_id_to_name = {"1": "FETCH"}
        cls.good_id_to_name = {"2": "Good_1"}
        cls.version_id = "v1"
        cls.list_of_messages = (
            DialogueMessage(
                TacMessage.Performative.REGISTER, {"agent_name": cls.agent_name}, True
            ),
            DialogueMessage(
                TacMessage.Performative.GAME_DATA,
                {
                    "amount_by_currency_id": cls.amount_by_currency_id,
                    "exchange_params_by_currency_id": cls.exchange_params_by_currency_id,
                    "quantities_by_good_id": cls.quantities_by_good_id,
                    "utility_params_by_good_id": cls.utility_params_by_good_id,
                    "fee_by_currency_id": cls.fee_by_currency_id,
                    "agent_addr_to_name": cls.agent_addr_to_name,
                    "currency_id_to_name": cls.currency_id_to_name,
                    "good_id_to_name": cls.good_id_to_name,
                    "version_id": cls.version_id,
                },
            ),
            DialogueMessage(
                TacMessage.Performative.TRANSACTION,
                {
                    "transaction_id": "some_transaction_id",
                    "ledger_id": "some_ledger_id",
                    "sender_address": "some_sender_address",
                    "counterparty_address": "some_counterparty_address",
                    "amount_by_currency_id": {"1": 5},
                    "fee_by_currency_id": {"1": 1},
                    "quantities_by_good_id": {"2": -5},
                    "nonce": "some_nonce",
                    "sender_signature": "some_sender_signature",
                    "counterparty_signature": "some_counterparty_signature",
                },
            ),
        )

        cls.list_of_state_update_messages = (
            DialogueMessage(
                StateUpdateMessage.Performative.INITIALIZE,
                {
                    "amount_by_currency_id": cls.amount_by_currency_id,
                    "quantities_by_good_id": cls.quantities_by_good_id,
                },
            ),
        )

        cls.game._expected_controller_addr = COUNTERPARTY_ADDRESS

    def test_setup(self):
        """Test the setup method of the tac handler."""
        assert self.tac_handler.setup() is None
        self.assert_quantity_in_outbox(0)

    def test_handle_sender_not_equal_to_expected_controller(self):
        """Test the handle method of the tac handler where message sender is NOT equal to the expected controller."""
        # setup
        self.game._expected_controller_addr = "some_different_controller_address"

        dialogue = self.prepare_skill_dialogue(
            self.tac_dialogues, self.list_of_messages[:1]
        )

        incoming_message = self.build_incoming_message_for_skill_dialogue(
            dialogue=dialogue,
            message_type=TacMessage,
            performative=TacMessage.Performative.TAC_ERROR,
            error_code=TacMessage.ErrorCode.AGENT_ADDR_ALREADY_REGISTERED,
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            with pytest.raises(
                ValueError,
                match="The sender of the message is not the controller agent we registered with.",
            ):
                self.tac_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.DEBUG,
            f"handling controller response. performative={incoming_message.performative}",
        )

    def test_handle_unidentified_dialogue(self):
        """Test the _handle_unidentified_dialogue method of the tac handler."""
        # setup
        incorrect_dialogue_reference = ("", "")
        incoming_message = self.build_incoming_message(
            message_type=TacMessage,
            dialogue_reference=incorrect_dialogue_reference,
            performative=TacMessage.Performative.CANCELLED,
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.tac_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.WARNING,
            f"received invalid tac message={incoming_message}, unidentified dialogue.",
        )

    def test_on_tac_error_code_i(self):
        """Test the _on_tac_error method of the tac handler where error_code is NOT TRANSACTION_NOT_VALID."""
        # setup
        dialogue = self.prepare_skill_dialogue(
            self.tac_dialogues, self.list_of_messages[:1]
        )

        error_code = TacMessage.ErrorCode.AGENT_ADDR_ALREADY_REGISTERED
        incoming_message = self.build_incoming_message_for_skill_dialogue(
            dialogue=dialogue,
            message_type=TacMessage,
            performative=TacMessage.Performative.TAC_ERROR,
            error_code=error_code,
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.tac_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.DEBUG,
            f"received error from the controller in dialogue={dialogue}. error_msg={TacMessage.ErrorCode.to_msg(error_code.value)}",
        )

    def test_on_tac_error_code_ii(self):
        """Test the _on_tac_error method of the tac handler where error_code is TRANSACTION_NOT_VALID."""
        # setup
        dialogue = self.prepare_skill_dialogue(
            self.tac_dialogues, self.list_of_messages[:1]
        )

        tx_id = "some_tx_id"
        error_code = TacMessage.ErrorCode.TRANSACTION_NOT_VALID
        incoming_message = self.build_incoming_message_for_skill_dialogue(
            dialogue=dialogue,
            message_type=TacMessage,
            performative=TacMessage.Performative.TAC_ERROR,
            error_code=error_code,
            info={"transaction_id": tx_id},
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.tac_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.DEBUG,
            f"received error from the controller in dialogue={dialogue}. error_msg={TacMessage.ErrorCode.to_msg(error_code.value)}",
        )
        mock_logger.assert_any_call(
            logging.WARNING, f"received error on transaction id: {tx_id[-10:]}",
        )

    def test_on_start_i(self):
        """Test the _on_start method of the tac handler."""
        # setup
        self.game._phase = Phase.GAME_REGISTRATION
        self.game._is_using_contract = False

        dialogue = self.prepare_skill_dialogue(
            self.tac_dialogues, self.list_of_messages[:1]
        )

        incoming_message = cast(
            TacMessage,
            self.build_incoming_message_for_skill_dialogue(
                dialogue=dialogue,
                message_type=TacMessage,
                performative=TacMessage.Performative.GAME_DATA,
                amount_by_currency_id=self.amount_by_currency_id,
                exchange_params_by_currency_id=self.exchange_params_by_currency_id,
                quantities_by_good_id=self.quantities_by_good_id,
                utility_params_by_good_id=self.utility_params_by_good_id,
                fee_by_currency_id=self.fee_by_currency_id,
                agent_addr_to_name=self.agent_addr_to_name,
                currency_id_to_name=self.currency_id_to_name,
                good_id_to_name=self.good_id_to_name,
                version_id=self.version_id,
            ),
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            with patch.object(self.game, "init"):
                self.tac_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO,
            "received start event from the controller. Starting to compete...",
        )

        self.assert_quantity_in_decision_making_queue(1)
        has_attributes, error_str = self.message_has_attributes(
            actual_message=self.get_message_from_decision_maker_inbox(),
            message_type=StateUpdateMessage,
            performative=StateUpdateMessage.Performative.INITIALIZE,
            to=self.skill.skill_context.decision_maker_address,
            sender=self.skill.skill_context.agent_address
            + "_"
            + str(self.skill.skill_context.skill_id),
            amount_by_currency_id=incoming_message.amount_by_currency_id,
            quantities_by_good_id=incoming_message.quantities_by_good_id,
            exchange_params_by_currency_id=incoming_message.exchange_params_by_currency_id,
            utility_params_by_good_id=incoming_message.utility_params_by_good_id,
        )
        assert has_attributes, error_str
        assert (
            self.skill.skill_context.shared_state["fee_by_currency_id"]
            == incoming_message.fee_by_currency_id
        )
        assert self.game.state_update_dialogue is not None

    def test_on_start_ii(self):
        """Test the _on_start method of the tac handler where phase is NOT GAME_REGISTRATION."""
        # setup
        self.game._phase = Phase.PRE_GAME
        self.game._is_using_contract = False

        dialogue = self.prepare_skill_dialogue(
            self.tac_dialogues, self.list_of_messages[:1]
        )

        incoming_message = self.build_incoming_message_for_skill_dialogue(
            dialogue=dialogue,
            message_type=TacMessage,
            performative=TacMessage.Performative.GAME_DATA,
            amount_by_currency_id=self.amount_by_currency_id,
            exchange_params_by_currency_id=self.exchange_params_by_currency_id,
            quantities_by_good_id=self.quantities_by_good_id,
            utility_params_by_good_id=self.utility_params_by_good_id,
            fee_by_currency_id=self.fee_by_currency_id,
            agent_addr_to_name=self.agent_addr_to_name,
            currency_id_to_name=self.currency_id_to_name,
            good_id_to_name=self.good_id_to_name,
            version_id=self.version_id,
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.tac_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.WARNING,
            f"we do not expect a start message in game phase={self.game.phase.value}",
        )

    def test_on_start_iii(self):
        """Test the _on_start method of the tac handler where game uses contract."""
        # setup
        self.game._phase = Phase.GAME_REGISTRATION
        self.game._is_using_contract = True

        dialogue = self.prepare_skill_dialogue(
            self.tac_dialogues, self.list_of_messages[:1]
        )

        contract_address = "some_contract_address"
        incoming_message = cast(
            TacMessage,
            self.build_incoming_message_for_skill_dialogue(
                dialogue=dialogue,
                message_type=TacMessage,
                performative=TacMessage.Performative.GAME_DATA,
                amount_by_currency_id=self.amount_by_currency_id,
                exchange_params_by_currency_id=self.exchange_params_by_currency_id,
                quantities_by_good_id=self.quantities_by_good_id,
                utility_params_by_good_id=self.utility_params_by_good_id,
                fee_by_currency_id=self.fee_by_currency_id,
                agent_addr_to_name=self.agent_addr_to_name,
                currency_id_to_name=self.currency_id_to_name,
                good_id_to_name=self.good_id_to_name,
                version_id=self.version_id,
                info={"contract_address": contract_address},
            ),
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            with patch.object(self.game, "init"):
                with patch.object(
                    self.tac_handler, "_update_ownership_and_preferences"
                ) as mocked_uoap:
                    self.tac_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO,
            "received start event from the controller. Starting to compete...",
        )
        assert self.game.contract_address == contract_address
        assert (
            self.skill.skill_context.shared_state["erc1155_contract_address"]
            == contract_address
        )
        mock_logger.assert_any_call(
            logging.INFO, f"received a contract address: {contract_address}",
        )
        mocked_uoap.assert_called_once()

    def test_on_start_iv(self):
        """Test the _on_start method of the tac handler where game uses contract."""
        # setup
        self.game._phase = Phase.GAME_REGISTRATION
        self.game._is_using_contract = True

        dialogue = self.prepare_skill_dialogue(
            self.tac_dialogues, self.list_of_messages[:1]
        )

        incoming_message = cast(
            TacMessage,
            self.build_incoming_message_for_skill_dialogue(
                dialogue=dialogue,
                message_type=TacMessage,
                performative=TacMessage.Performative.GAME_DATA,
                amount_by_currency_id=self.amount_by_currency_id,
                exchange_params_by_currency_id=self.exchange_params_by_currency_id,
                quantities_by_good_id=self.quantities_by_good_id,
                utility_params_by_good_id=self.utility_params_by_good_id,
                fee_by_currency_id=self.fee_by_currency_id,
                agent_addr_to_name=self.agent_addr_to_name,
                currency_id_to_name=self.currency_id_to_name,
                good_id_to_name=self.good_id_to_name,
                version_id=self.version_id,
                info={"contract_address": None},
            ),
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            with patch.object(self.game, "init"):
                with patch.object(
                    self.tac_handler, "_update_ownership_and_preferences"
                ) as mocked_uoap:
                    self.tac_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO,
            "received start event from the controller. Starting to compete...",
        )
        mock_logger.assert_any_call(
            logging.WARNING, "did not receive a contract address!",
        )
        mocked_uoap.assert_not_called()

    def test_on_cancelled_i(self):
        """Test the _on_cancelled method of the tac handler where phase is GAME_REGISTRATION."""
        # setup
        self.game._phase = Phase.GAME_REGISTRATION

        dialogue = self.prepare_skill_dialogue(
            self.tac_dialogues, self.list_of_messages[:1]
        )

        incoming_message = cast(
            TacMessage,
            self.build_incoming_message_for_skill_dialogue(
                dialogue=dialogue,
                message_type=TacMessage,
                performative=TacMessage.Performative.CANCELLED,
            ),
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.tac_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO, "received cancellation from the controller.",
        )
        assert self.game.phase == Phase.POST_GAME
        assert self.skill.skill_context.is_active is False
        assert self.skill.skill_context.shared_state["is_game_finished"] is True

    def test_on_cancelled_ii(self):
        """Test the _on_cancelled method of the tac handler where phase is GAME."""
        # setup
        self.game._phase = Phase.GAME

        dialogue = self.prepare_skill_dialogue(
            self.tac_dialogues, self.list_of_messages[:1]
        )

        incoming_message = cast(
            TacMessage,
            self.build_incoming_message_for_skill_dialogue(
                dialogue=dialogue,
                message_type=TacMessage,
                performative=TacMessage.Performative.CANCELLED,
            ),
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.tac_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO, "received cancellation from the controller.",
        )
        assert self.game.phase == Phase.POST_GAME
        assert self.skill.skill_context.is_active is False
        assert self.skill.skill_context.shared_state["is_game_finished"] is True

    def test_on_cancelled_iii(self):
        """Test the _on_cancelled method of the tac handler where phase is NOT GAME_REGISTRATION nor GAME."""
        # setup
        self.game._phase = Phase.PRE_GAME

        dialogue = self.prepare_skill_dialogue(
            self.tac_dialogues, self.list_of_messages[:1]
        )

        incoming_message = cast(
            TacMessage,
            self.build_incoming_message_for_skill_dialogue(
                dialogue=dialogue,
                message_type=TacMessage,
                performative=TacMessage.Performative.CANCELLED,
            ),
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.tac_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.WARNING,
            f"we do not expect a message in game phase={self.game.phase.value}, received msg={incoming_message}",
        )

    def test_on_transaction_confirmed_i(self):
        """Test the _on_transaction_confirmed method of the tac handler."""
        # setup
        self.game._phase = Phase.GAME

        state_update_dialogue = cast(
            StateUpdateDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.state_update_dialogues,
                messages=self.list_of_state_update_messages[:1],
                counterparty=self.skill.skill_context.decision_maker_address,
            ),
        )
        self.game._state_update_dialogue = state_update_dialogue

        dialogue = self.prepare_skill_dialogue(
            self.tac_dialogues, self.list_of_messages[:3]
        )

        transaction_id = "some_transaction_id"
        incoming_message = cast(
            TacMessage,
            self.build_incoming_message_for_skill_dialogue(
                dialogue=dialogue,
                message_type=TacMessage,
                performative=TacMessage.Performative.TRANSACTION_CONFIRMATION,
                transaction_id=transaction_id,
                amount_by_currency_id=self.amount_by_currency_id,
                quantities_by_good_id=self.quantities_by_good_id,
            ),
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.tac_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO,
            f"received transaction confirmation from the controller: transaction_id={transaction_id}",
        )

        self.assert_quantity_in_decision_making_queue(1)
        has_attributes, error_str = self.message_has_attributes(
            actual_message=self.get_message_from_decision_maker_inbox(),
            message_type=StateUpdateMessage,
            performative=StateUpdateMessage.Performative.APPLY,
            to=self.skill.skill_context.decision_maker_address,
            sender=self.skill.skill_context.agent_address
            + "_"
            + str(self.skill.skill_context.skill_id),
            amount_by_currency_id=incoming_message.amount_by_currency_id,
            quantities_by_good_id=incoming_message.quantities_by_good_id,
        )
        assert has_attributes, error_str
        assert (
            incoming_message.transaction_id
            in self.skill.skill_context.shared_state["confirmed_tx_ids"]
        )

    def test_on_transaction_confirmed_ii(self):
        """Test the _on_transaction_confirmed method of the tac handler where phase is not GAME."""
        # setup
        self.game._phase = Phase.PRE_GAME

        dialogue = self.prepare_skill_dialogue(
            self.tac_dialogues, self.list_of_messages[:3]
        )

        transaction_id = "some_transaction_id"
        incoming_message = cast(
            TacMessage,
            self.build_incoming_message_for_skill_dialogue(
                dialogue=dialogue,
                message_type=TacMessage,
                performative=TacMessage.Performative.TRANSACTION_CONFIRMATION,
                transaction_id=transaction_id,
                amount_by_currency_id=self.amount_by_currency_id,
                quantities_by_good_id=self.quantities_by_good_id,
            ),
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.tac_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.WARNING,
            f"we do not expect a transaction in game phase={self.game.phase.value}, received msg={incoming_message}",
        )

    def test_on_transaction_confirmed_iii(self):
        """Test the _on_transaction_confirmed method of the tac handler where state_update dialogue is empty."""
        # setup
        self.game._phase = Phase.GAME

        state_update_dialogue = cast(
            StateUpdateDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.state_update_dialogues,
                messages=self.list_of_state_update_messages[:1],
                counterparty=self.skill.skill_context.decision_maker_address,
            ),
        )
        state_update_dialogue._incoming_messages = []
        state_update_dialogue._outgoing_messages = []
        self.game._state_update_dialogue = state_update_dialogue

        dialogue = self.prepare_skill_dialogue(
            self.tac_dialogues, self.list_of_messages[:3]
        )

        transaction_id = "some_transaction_id"
        incoming_message = cast(
            TacMessage,
            self.build_incoming_message_for_skill_dialogue(
                dialogue=dialogue,
                message_type=TacMessage,
                performative=TacMessage.Performative.TRANSACTION_CONFIRMATION,
                transaction_id=transaction_id,
                amount_by_currency_id=self.amount_by_currency_id,
                quantities_by_good_id=self.quantities_by_good_id,
            ),
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            with pytest.raises(ValueError, match="Could not retrieve last message."):
                self.tac_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO,
            f"received transaction confirmation from the controller: transaction_id={transaction_id}",
        )

    def test_handle_invalid(self):
        """Test the _handle_invalid method of the tac handler."""
        # setup
        tac_dialogue = self.prepare_skill_dialogue(
            dialogues=self.tac_dialogues, messages=self.list_of_messages[:1],
        )
        incoming_message = self.build_incoming_message_for_skill_dialogue(
            dialogue=tac_dialogue,
            performative=TacMessage.Performative.UNREGISTER,
            sender=COUNTERPARTY_ADDRESS,
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.tac_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.WARNING,
            f"cannot handle tac message of performative={incoming_message.performative} in dialogue={tac_dialogue}.",
        )

    def test_teardown(self):
        """Test the teardown method of the tac handler."""
        assert self.tac_handler.teardown() is None
        self.assert_quantity_in_outbox(0)
